MESSAGE("DownloadBoost starting...")

SET(BOOST_LOCAL_PATH ${ROOT_DIR}/external_dependencies/boost)

INCLUDE(FetchContent)

##################
# Download Boost #
##################
SET(WINAPI_BOOST_LIB "")
IF(WIN32)
    SET(WINAPI_BOOST_LIB libs/winapi)
ENDIF()

FETCHCONTENT_DECLARE(
    boost
    SOURCE_DIR ${BOOST_LOCAL_PATH}
    GIT_REPOSITORY https://github.com/boostorg/boost.git
    GIT_SUBMODULES tools/build tools/boost_install
    libs/system libs/random libs/serialization
    libs/config libs/headers libs/assert libs/core
    libs/integer libs/type_traits libs/mpl
    libs/preprocessor libs/throw_exception
    libs/utility libs/static_assert libs/smart_ptr
    libs/predef libs/move libs/io libs/iterator
    libs/detail libs/spirit libs/optional
    libs/function libs/type_index libs/bind
    libs/container_hash libs/array libs/test
    libs/timer libs/exception libs/algorithm
    libs/range libs/numeric libs/format
    libs/lexical_cast libs/concept_check
    libs/container libs/math libs/function_types
    libs/typeof ${WINAPI_BOOST_LIB}
)

SET(FETCHCONTENT_QUIET FALSE)

FETCHCONTENT_POPULATE(boost)

###############
# Build Boost #
###############
SET(BOOTSTRAP_CMD sh bootstrap.sh)
SET(B2_CMD ./b2 -j${N_CORES})
IF(WIN32)
    SET(BOOTSTRAP_CMD bootstrap.bat)
    SET(B2_CMD b2 -j${N_CORES})
ENDIF()

EXECUTE_PROCESS(
    COMMAND ${BOOTSTRAP_CMD}
    WORKING_DIRECTORY ${BOOST_LOCAL_PATH}
    RESULT_VARIABLE rv
)
IF(NOT rv STREQUAL "0")
    MESSAGE("Boost build: bootstrap: ${rv}")
ENDIF()

EXECUTE_PROCESS(
    COMMAND ${B2_CMD} headers
    WORKING_DIRECTORY ${BOOST_LOCAL_PATH}
    RESULT_VARIABLE rv
)
IF(NOT rv STREQUAL "0")
    MESSAGE("Boost build: b2 headers: ${rv}")
ENDIF()

SET(VARIANT release)
IF(${CMAKE_BUILD_TYPE} STREQUAL "Debug")
    SET(VARIANT debug)
ENDIF()

set(PIC_CODE "")
if(NOT WIN32)
    set(PIC_CODE cxxflags=-fPIC)
endif()

EXECUTE_PROCESS(
    COMMAND ${B2_CMD} -q ${PIC_CODE} ${BOOST_TOOLSET} --layout=system variant=${VARIANT} link=${DEPENDENCIES_LINK_TYPE} address-model=64 architecture=x86 --with-system --with-serialization --with-random
    WORKING_DIRECTORY ${BOOST_LOCAL_PATH}
    RESULT_VARIABLE rv
)
IF(NOT rv STREQUAL "0")
    MESSAGE("Boost build: b2: ${rv}")
ENDIF()

FIND_PACKAGE(
    Boost

    COMPONENTS
    system
    serialization
    random
)

IF(NOT Boost_FOUND)
    MESSAGE(FATAL_ERROR "Boost was not downloaded successfully!")
ENDIF()

