MESSAGE("DownloadOpenBLAS starting...")

SET(OPENBLAS_LOCAL_PATH ${ROOT_DIR}/external_dependencies/OpenBLAS)

INCLUDE(FetchContent)

#####################
# Download OpenBLAS #
#####################
IF(WIN32)
    FETCHCONTENT_DECLARE(
        OpenBLAS
        SOURCE_DIR ${OPENBLAS_LOCAL_PATH}
        URL https://github.com/JuliaLinearAlgebra/OpenBLASBuilder/releases/download/v0.3.0-3/OpenBLAS.v0.3.0.x86_64-w64-mingw32-gcc8.tar.gz
    )

    # See messages during 'git clone'
    SET(FETCHCONTENT_QUIET FALSE)

    IF(NOT OpenBLAS_POPULATED)
        FETCHCONTENT_POPULATE(OpenBLAS)
    ENDIF()

ELSE()
    FETCHCONTENT_DECLARE(
        OpenBLAS
        SOURCE_DIR ${OPENBLAS_LOCAL_PATH}
        GIT_REPOSITORY https://github.com/xianyi/OpenBLAS.git
    )

    # See messages during 'git clone'
    SET(FETCHCONTENT_QUIET FALSE)

    IF(NOT OpenBLAS_POPULATED)
        FETCHCONTENT_POPULATE(OpenBLAS)
    ENDIF()
ENDIF()

IF(NOT WIN32)
    ###################################
    # Build OpenBLAS (only for Linux) #
    ###################################
    EXECUTE_PROCESS(
        COMMAND ${CMAKE_COMMAND} -j ${N_CORES} .
        WORKING_DIRECTORY ${OPENBLAS_LOCAL_PATH}
        RESULT_VARIABLE rv
    )
    IF(NOT rv STREQUAL "0")
        MESSAGE("OpenBLAS build: cmake .: ${rv}")
    ENDIF()

    # Build library
    EXECUTE_PROCESS(
        COMMAND ${CMAKE_COMMAND} --build . --config Release -j ${N_CORES}
        WORKING_DIRECTORY ${OPENBLAS_LOCAL_PATH}
        RESULT_VARIABLE rv
    )
    IF(NOT rv STREQUAL "0")
        MESSAGE("OpenBLAS build: cmake --build: ${rv}")
    ENDIF()
ENDIF()

FIND_PACKAGE(OpenBLAS)
