# This file is NOT an original FindBoost.cmake module provided by KitWare!
#
# It's a simplified version whose only purpose is to be used in a software
# library lib4neuro and it does NOT provide a full funcionality of the original,
# as it only works with the system-layout-named libraries (e.g. libboost_system.so).

# Optional user-defined variables
# (can be passed directly to CMake or exported as environmental variables
#
# * BOOST_LIBRARYDIR - The path to the folder containing Boost libraries
# * BOOST_INCLUDEDIR - The path to the folder containing Boost header files

# "External" used variables
#
# * DEPENDENCIES_LINK_TYPE - TODO

# "Output" variables set after running this script
# * Boost_FOUND - TODO
# * Boost_INCLUDE_DIRS - TODO
# * Boost_LIBRARY_DIRS - TODO
# * Boost_LIBRARIES - TODO

# Module usage
# TODO

MESSAGE("FindBoost starting...")

# Check if needed Boost components were specified
IF(NOT Boost_FIND_COMPONENTS)
    MESSAGE(FATAL_ERROR "No Boost components were specified! Please, set them correctly with flag COMPONENTS (see Module Usage section in this script).")
ELSE()
    MESSAGE("Required Boost components: ${Boost_FIND_COMPONENTS}")
ENDIF()

# Look for a standard boost header file.
SET(Boost_INCLUDE_DIRS "Boost_INCLUDE_DIRS-NOTFOUND")
FIND_PATH(
    Boost_INCLUDE_DIRS

    NAMES
    config.hpp

    HINTS
    $ENV{BOOST_ROOT}
    $ENV{BOOST_ROOT}/boost
    ${BOOST_INCLUDEDIR}
    $ENV{BOOST_INCLUDEDIR}
    ${ROOT_DIR}/external_dependencies/boost
    ${ROOT_DIR}/external_dependencies/boost/src/boost

    PATHS
    /usr/include

    PATH_SUFFIXES
    boost
    include
)

# Add path without "boost" include sub-directory to include path,
# as Boost headers are supposed to be included like
# #include<boost/...> according to the documentation
SET(TMP "")
STRING(REPLACE "/boost/boost" "/boost" TMP ${Boost_INCLUDE_DIRS})
LIST(APPEND Boost_INCLUDE_DIRS ${TMP})

IF(NOT Boost_INCLUDE_DIRS)
    #    message(FATAL_ERROR "Boost include directory was not found! Please, set variable BOOST_INCLUDEDIR to the correct path.")
ELSE()
    MESSAGE("Boost_INCLUDE_DIRS: ${Boost_INCLUDE_DIRS}")
ENDIF()

# Create a list of requested Boost libraries with "system" names
IF(NOT DEPENDENCIES_LINK_TYPE)
    MESSAGE(FATAL_ERROR "Variable DEPENDENCIES_LINK_TYPE is not set! Set it to 'static' or 'shared'.")
ENDIF()

SET(REQUESTED_BOOST_LIBS "")
FOREACH(COMPONENT ${Boost_FIND_COMPONENTS})
    LIST(APPEND REQUESTED_BOOST_LIBS "${LIB_PREFIX}boost_${COMPONENT}.${LIB_SUFFIX}")
ENDFOREACH()

MESSAGE("REQUESTED_BOOST_LIBS: ${REQUESTED_BOOST_LIBS}")

# Look for libraries specified by COMPONENTS flag
SET(Boost_LIBRARY_DIRS "Boost_LIBRARY_DIRS-NOTFOUND")
FIND_PATH(
    Boost_LIBRARY_DIRS

    NAMES
    ${REQUESTED_BOOST_LIBS}

    HINTS
    $ENV{BOOST_ROOT}
    $ENV{BOOST_ROOT}/stage
    $ENV{BOOST_ROOT}/stage/lib
    ${BOOST_LIBRARYDIR}
    $ENV{BOOST_LIBRARYDIR}
    ${ROOT_DIR}/external_dependencies/boost
    ${ROOT_DIR}/external_dependencies/boost/src/boost
    ${ROOT_DIR}/external_dependencies/boost/stage
    ${ROOT_DIR}/external_dependencies/boost/stage/lib


    PATHS
    /usr/lib/boost
    /usr/lib/x86_64-linux-gnu

    PATH_SUFFIXES
    lib
)

IF(NOT Boost_LIBRARY_DIRS)
    #message(FATAL_ERROR "Boost library directory was not found! Please, set variable BOOST_LIBRARYDIR to the correct path.")
ELSE()
    MESSAGE("Boost_LIBRARY_DIRS: ${Boost_LIBRARY_DIRS}")

    # Construct list of libraries' names and make them
    # targets, so they may be linked
    SET(Boost_LIBRARIES "")
    FOREACH(LIBNAME ${REQUESTED_BOOST_LIBS})
        MESSAGE("Looking for ${LIBNAME}...")

        SET(${LIBNAME} "${LIBNAME}-NOTFOUND")
        FIND_LIBRARY(
            ${LIBNAME}

            NAMES
            ${LIBNAME}

            PATHS
            ${Boost_LIBRARY_DIRS}

            PATH_SUFFIXES
            stage/lib
            lib

            NO_DEFAULT_PATH
        )

        # Check, if the Boost component was found
        IF("${${LIBNAME}}" STREQUAL "${LIBNAME}-NOTFOUND")
            #message(FATAL_ERROR "Boost library ${LIBNAME} was NOT found!\
            #                     Please, set variable BOOST_LIBRARYDIR to the correct path and check the library names\
            #                     format in your Boost installation.")
        ELSE()
            MESSAGE("${LIBNAME} was found: ${${LIBNAME}}")

            # Add every found library as an IMPORTED target
            STRING(TOUPPER ${DEPENDENCIES_LINK_TYPE} TMP)
            STRING(REGEX REPLACE "^lib" "" TARGET_NAME ${LIBNAME})
            STRING(REGEX REPLACE "\\.[a-z]*$" "" TARGET_NAME ${TARGET_NAME})
            ADD_LIBRARY(${TARGET_NAME} ${TMP} IMPORTED)
            SET_TARGET_PROPERTIES(${TARGET_NAME} PROPERTIES IMPORTED_LOCATION ${${LIBNAME}})
            MESSAGE("Created IMPORTED library target: ${TARGET_NAME}")
        ENDIF()

        LIST(APPEND Boost_LIBRARIES ${${LIBNAME}})
    ENDFOREACH()

ENDIF()


# Set Boost_FOUND
INCLUDE(FindPackageHandleStandardArgs)
FIND_PACKAGE_HANDLE_STANDARD_ARGS(
    Boost

    FAIL_MESSAGE
    "Boost was NOT found!"

    REQUIRED_VARS
    Boost_INCLUDE_DIRS
    Boost_LIBRARY_DIRS
)

IF(Boost_FOUND)
    MESSAGE("Boost_INCLUDE_DIRS: ${Boost_INCLUDE_DIRS}")
    MESSAGE("Boost_LIBRARY_DIRS: ${Boost_LIBRARY_DIRS}")
    MESSAGE("Boost_LIBRARIES: ${Boost_LIBRARIES}")
ENDIF()
