/**
 * DESCRIPTION OF THE CLASS
 *
 * @author David Vojtek
 * @date 2018
 */

#define BOOST_TEST_MODULE Particle_test

// TODO fix boost_test_lib and remove the following include!
#ifdef _WINDOWS
#include <boost/test/included/unit_test.hpp>
#endif

#include "boost_unit_tests_preamble.h"
#include "../ErrorFunction/ErrorFunctionsMock.h"
#include "../LearningMethods/ParticleSwarm.h"

using namespace lib4neuro;

/**
 * Boost testing suite for testing ParticleSwarm.h
 * TODO
 */
BOOST_AUTO_TEST_SUITE(Particle_test)

    BOOST_AUTO_TEST_CASE(Particle_construction_test) {
        std::vector<double> domain_bound{1, 2, 3, 4, 5};
        mock_ErrorFunction  error;
        MOCK_EXPECT(error.get_dimension).once().returns(5);
        MOCK_EXPECT(error.eval).once().returns(0.8);
        BOOST_CHECK_NO_THROW(Particle(&error,
                                      &domain_bound));
    }

    BOOST_AUTO_TEST_CASE(Particle_get_coordinate_test) {
        std::vector<double> domain_bound{1, 2, 3, 4, 5};
        mock_ErrorFunction  error;

        MOCK_EXPECT(error.get_dimension).returns(5);
        MOCK_EXPECT(error.eval).returns(0.8);

        Particle particle1(&error,
                           &domain_bound);
        Particle particle2(&error,
                           &domain_bound);

        BOOST_CHECK(*particle1.get_coordinate() != *particle2.get_coordinate());
    }

    BOOST_AUTO_TEST_CASE(Particle_get_optimal_value_test) {
        std::vector<double> domain_bound{1, 2, 3, 4, 5};
        mock_ErrorFunction  error;

        MOCK_EXPECT(error.get_dimension).returns(5);
        MOCK_EXPECT(error.eval).returns(0.8);


        Particle particle1(&error,
                           &domain_bound);
        BOOST_CHECK_EQUAL(0.8,
                          particle1.get_optimal_value());
    }


BOOST_AUTO_TEST_SUITE_END()
